// --- [ alarm.c ] ------------------------------------------------------------
//      30.11.09
// Modul fr alle zeitgesteuerten Schaltvorgnge

#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>

#include "rs232.h"
#include "global.h"
#include "lcd.h"
#include "sun.h"
#include "dcf_bits.h"
#include "TWI_master_slave.h"
#include "dcf_77.h"

uint8_t schalter[2][9];						// enthlt die Schalterstellungen (alt und neu)
													// fr 9 * 8 Schalter (8 TWI-Slaves und bis zu 8 Portpins)
extern uint8_t error;

#ifdef ZEIT_ZEICHEN
// -----------------------------------------------------------------------------
// sendet jeweils zur vollen Viertelstunde ein Signal an einen TWI-Slave
void zeit_zeichen(void)
{
uint8_t i;

if (time[CLK][STUNDE] < 6) return;			// Nachtruhe zwischen 0 und 6 Uhr !

twi_msg[0] = TWI_ADR_ZEIT_ZEICHEN;			// globale Variable

if (time[CLK][MINUTE] == 0)
	{
	i = (time[CLK][STUNDE]);
	if (i > 12) i -= 12;							// wenn Stunden > 12, dann bei 1 beginnen
	else if (i == 0) i = 12;
	twi_msg[1] = (i + 3);						// sendet (3 + Anzahl der Stunden)
	}

else if (time[CLK][MINUTE] == 15)
	{
	twi_msg[1] = 1;
	}

else if (time[CLK][MINUTE] == 30)
	{
	twi_msg[1] = 2;
	}

else if (time[CLK][MINUTE] == 45)
	{
	twi_msg[1] = 3;
	}

else
	{
	return;
	}

TWI_MA_Start_Transceiver_With_Data(twi_msg, 2);
TWI_SLA_restart();

}
#endif


// -----------------------------------------------------------------------------
// vergleicht die aktuelle Uhr-Zeit mit den Zeiten fr Sonnenauf- und Untergang.
// Bei Gleichheit (und somit nur jeweils einmal am Tag) wird ein Switch ein-/ausgeschaltet.
// Funktion: bei Sonnenuntergang wird Licht eingeschaltet und
// bei Sonnenaufgang abgeschaltet.
// Unabhngig von dieser Funktion kann auf dieser Switch auch via Schaltdefinition gendert werden.
#ifdef DAY_LIGHT
void DCF_check_daylight(void)
{
													// Sonnenaufgang
if ((time[CLK][STUNDE] == time[SUN][SUN_UP_H]) && (time[CLK][MINUTE] == time[SUN][SUN_UP_M]))
	{
													// switch lschen ("festverdrahtet" auf PD.0)
	ALARMPORT &= (~(1<<DAYLIGHT_PIN));

#ifdef LOG_RS232
	print_datum();
	puts_p(PSTR("Sonnenaufgang"));
#endif

	}
													// Sonnenuntergang
else if ((time[CLK][STUNDE] == time[SUN][SUN_DN_H]) && (time[CLK][MINUTE] == time[SUN][SUN_DN_M]))
	{
													// switch setzen ("festverdrahtet" auf PD.0)
	ALARMPORT |= 1<<DAYLIGHT_PIN;
#ifdef LOG_RS232
	print_datum();
	puts_p(PSTR("Sonnenuntergang"));
#endif

	}
}
#endif

// -----------------------------------------------------------------------------
// vergleicht die Bytes der Schalt-Arrays mit time[CLK][],
// sofern Bit.7 in time[][x] gesetzt ist, wird genau dieses Byte nicht verglichen !
// Wenn der Vergleich erfolgreich ist, wird die hinterlegte Aktion ausgefhrt
// (also ein Portpin geschaltet oder an einen TWI-Slave eine Message gesendet)
void DCF_check_alarm(void)
{
uint8_t i, j;
uint8_t r;
uint8_t twi_adr[] = {TWI1, TWI2, TWI3, TWI4, TWI5, TWI6, TWI7, TWI8}; // TWI-Adressen

// die Schaltdefinitionen liegen hinter den Uhr-internen-Arrays ...
for (i = CLK_ARRAYS; i < (ALARMDEFS + CLK_ARRAYS); i++)
	{
	if (!(time[i][MODUS])) continue;		// wenn die Alarmdefinition (das letzte Feld im Array) leer ist
													// in MODUS steht, was der Alarm bewirken soll
	r = TRUE;
	j = 6;										// Element mit Index 5 ... 0 vergleichen

	do {
		j--;
		if (!(time[i][j] & 0x80))			// kein Vergleich, wenn mindestens Bit.7 gesetzt ist
													// damit kann nicht nach Sommerzeit gefiltert werden,
													// sondern nur nach Wochentagen
	//	if (!(time[i][j] == 128))			// kein Vergleich, wenn genau      Bit.7 gesetzt ist
													// dadurch kann auch die Sommerzeit als Bit.7 in DOW
													// als Suchkriterium dienen - ggf. in Verbindung mit der
													// Angabe eines Wochentages.
													// Irritierend ist dabei aber, dass sich nun DOW nicht mehr
													// im Rahmen der blichen Regel (Bit.7 = 1) vom Vergleich ausschlieen lsst.
													// Behelfsmig wrde der Wert '0' den Vergleich definitiv scheitern lassen.

			{										// ein einziger fehlerhafter Vergleich reicht aus
			if (j == DOW)						// beim Wochentag wird die Bitmaske verglichen
				{
				if (!(time[CLK][DOW] & time[i][DOW])) r = FALSE;
				}
			else									// bei allen anderen Daten die bereinstimmung
				{
				if (! (time[i][j] == time[CLK][j]) ) r = FALSE;
				}
			}
		} while (j && r);						// solange j > 0 und r = TRUE

	if (r) 										// wenn alle Vergleiche erfolgreich
		{
		DCF_set_alarm(i);						// dann ist ein Alarmzeitpunkt erreicht

#ifdef LOG_RS232
		print_datum();							// Neue Zeile und aktuelle Schaltzeit ausgeben
		print_alarm(i);						// die Daten des Alarms via RS232 ausgeben
#endif

		if (time[i][MODUS] & 0x80)			// wenn Bit.7 gesetzt, dann war es ein einmaliger Alarm
			{										// dann alle Alarmeintrge lschen
			for (j = 0; j < TIMEBYTES; j++) time[i][j] = 0;
			}
		}
	}

// nun sind alle SChaltdefinitionen durchsucht und alle Schaltaufgaben in switch[NEW_STATE][]
// geschreiben.
// Es ist jetzt festzustellen, ob sich Vernderungen gegenber dem letzten Zustand,
// der in switch[LAST_STATE][] steht, ergeben haben.
// Wenn ja, dann wird der Slave oder der Port aktualisiert.


for (j = 0; j < 9; j++)
	{												// wenn NEW_STATE ungleich LAST_STATE
	if (!(schalter[LAST_STATE][j] == schalter[NEW_STATE][j]))
		{											// LAST_STATE aktualisieren
		schalter[LAST_STATE][j] = schalter[NEW_STATE][j];
													// die TWI-Adresse holen
		if (j < 8)								// dann wird ein Portexpander angesprochen
			{
			twi_msg[0] = twi_adr[j];
			twi_msg[1] = schalter[NEW_STATE][j];// die Schalterstellung
			TWI_MA_Start_Transceiver_With_Data(twi_msg, 2);
			TWI_SLA_restart();				// warten bis die Sendung raus ist
			}

		else										// j == 8 -> Ausgabe auf PORTPIN -> siehe ALARMPINS = 64 !

			{
			// es sind an PortD die Pins 0,1,4,5,6,7 frei.
			// die Bits 2..5 mssen nach links auf Bit 4 .. 7 verschoben werden,
			// damit sie zu den freien Pins passen.
			i = schalter[NEW_STATE][j]    & 0b00000001;   // nur Schalter 64 bercksichtigen
//			i += ((schalter[NEW_STATE][j] & 0x00111100) << 2);


			if (i & (1<< PD0))				// es wird PD.7 wird genutzt
				{
				PORTD |= (1<<PD7);			// PD.7 setzen
				}
			else
				{
				PORTD &= (~(1<<PD7));		// PD.7 lschen
				}
			}
		}
	}

return;
}

// -----------------------------------------------------------------------------
// es ist ein Alarmzeitpunkt erreicht
// im Parameter idx steht, welches Alarmarray zu bearbeiten ist.
// im Modus  :		0 = inaktiv
//						1 = einschalten
//						2 = ausschalten
//						3 = togglen
//					  >3 = ein Byte an einen TWI-Slave senden
//                    die Schalter-Nr. wird als TWI-Adresse gewhlt
//                    die Bits 3-6 sind die Message, die gesendet wird.
//                    (das Bit.7 fehlt in der Message, weil es eine besondere Funktion hat)!
//
// die adressierbaren TWI_Schalter sind von 0 .. 63 durchnummeriert.
// Als Portpins am mc sind die Schalternr. 64 .. 69 vorbereitet.
// die letzte Schalterstellung steht in switch[LAST_STATE][].
// nderungen werden geschrieben     in switch[NEW_STATE][].
// wenn sich Vernderungen zwischen LAST_STATE und NEW_STATE ergeben,
// dann wird eine Nachricht an den Portexpander gesendet.
//
// in Schalter:	0  .. 63 wird auf TWI ausgegeben
//                64 .. 69 wird auf PORTD ausgegeben (aber z.Z. ist nur PD.7 als #64 aktiv)

void DCF_set_alarm(uint8_t idx)
{

uint8_t s, m;
uint8_t array_idx, pin_idx;
s = time[idx][TIMEBYTES - 2];				// Schalter: vorletztes Element im Array

m = time[idx][TIMEBYTES - 1];				// Modus   :    letztes Element im Array
m &= 0x7F;										// Bit.7 maskieren = Flag fr einmaligen Alarm

if (m < 4)										// fr Modus == 1, 2, 3
	{
	if (s > ALARMPINS) return;				// bei undefinierte Schalternummer beenden

	array_idx = s / 8;						// Index fr schalter[][] bestimmen
	pin_idx   = s % 8;						// Bit Nr.  bestimmen
	s = 1<< pin_idx;							// aus Bit Nr. eine Bitmaske berechnen

	if      (m == 1) schalter[NEW_STATE][array_idx] |=  s;	// Bit auf high legen
	else if (m == 2) schalter[NEW_STATE][array_idx] &= ~s;	// Bit auf low  legen
	else if (m == 3) schalter[NEW_STATE][array_idx] ^=  s;	// Bit togglen
	}

else 												// Modus  4 ... 127 -> ein Kommado an einen Slave senden
	{												// der Wert von Modus wird gesendet
	twi_msg[0] = (s & 0xFE);				// der Schalter ist in diesem Falle die TWI-Adresse,
													// bei der Bit.0 maskiert wird
													// (es soll ja geschrieben und nicht gelesen werden!)
													// Bits 2,3,4,5,6 -> die message
													// Bits 0,1,      -> was ist zu tun
													// Bit  7         -> schaltdefinition nach Ausfhrung lschen
	twi_msg[1] = m; 							// Bit.7 wurde bereits oben gelscht, daher Wertebereich 4 .. 127
	TWI_MA_Start_Transceiver_With_Data( twi_msg, 2);
	TWI_SLA_restart();
	}
}


// -----------------------------------------------------------------------------
// testet, ob ein Schaltzeitpunkt definiert ist - und wenn ja, wieviele.
// das letzte Byte eines Alarmslots legt fest, was zu tun ist.
// Ist der Alarm inaktiv, ist dieses Byte == 0;

#ifdef LCD
void DCF_show_alarm(void)
{
static uint8_t alarms = 0;						// merkt sich die Anzahl der aktiven Alarmdefinitionen
uint8_t i, j = 0;

for (i = CLK_ARRAYS; i < (ALARMDEFS + CLK_ARRAYS); i++)
	{
	if (time[i][MODUS]) j++;
	}
i = 0;												// Zhler zurcksetzen
lcd_goto_xy(2,1);

if (!(j == alarms))								// Anzeige/Ausgabe nur, wenn der Wert verndert wurde
	{
	alarms = j;
	if (j)
		{
		lcd_puts_p(PSTR("a:_"));
		lcd_puts(utoa(j,msg,10));
		lcd_putc('_');								// damit ggf. die zweite Stelle gelscht wird
														// beim Wechsel von zweistellig auf einstellig (10 auf 9)
		}

	else												// wenn kein Alarm mehr definiert
		{
		lcd_puts_p(PSTR("_____"));				// alte Meldung mit Unterstrich "lschen"
		}
	}
}
#endif


// ----------------------------------------------------------------------------
// Schreibt den Bereich der Schaltdefinitionen ins Eeprom,
// gleichgltig, ob eine explizite Definition vorliegt

void DCF_save_alarm(void)
{
uint16_t eeprom_adr = EEPROM_START;
uint8_t i, j;

// die Schaltdefinitionen liegen hinter den Zeit-Arrays !
for (i = CLK_ARRAYS; i < (ALARMDEFS + CLK_ARRAYS); i++)
	{

	for (j = 0; j < TIMEBYTES; j++)
		{
		while (EECR & (1<<EEPE));
		EEARH = (uint8_t) (eeprom_adr / 256);// Adresse schreiben
		EEARL = (uint8_t) eeprom_adr++;

		EEDR  = time[i][j];						// Daten schreiben

		cli();
		EECR |= (1<<EEMWE);
		EECR |= (1<<EEWE);
		sei();

		}
	}
EEARH = 0;											// gegen potentielle Zerstrung des Eeprom-Inhaltes
EEARL = 0;
}


// ----------------------------------------------------------------------------
// Liest alle Alarmdefinitionen aus dem Eeprom, wenn denn welche gesichert waren
// bricht ab, sobald ein 0xFF angetroffen wird !!
void DCF_read_alarm(void)
{
uint16_t eeprom_adr = EEPROM_START;			// Ab Adresse 8 im Eeprom einlesen
uint8_t i, j;

// die Schaltdefinitionen liegen hinter den Zeit-Arrays !
for (i = CLK_ARRAYS; i < (ALARMDEFS + CLK_ARRAYS); i++)
	{

	for (j = 0; j < TIMEBYTES; j++)
		{
		while (EECR & (1<<EEPE));
		EEARH = (uint8_t) (eeprom_adr / 256);	// Adresse schreiben
		EEARL = (uint8_t) eeprom_adr++;
		EECR |= (1<<EERE);						// Eeprom lesen
		if ((EEDR == 0xFF) && (!(j))) return;// wenn EEDR == 255 und j == 0
		time[i][j] = EEDR;						// Daten speichern
		}
	}
EEARH = 0;											// gegen potentielle Zerstrung des Eeprom-Inhaltes
EEARL = 0;
}

// ----------------------------------------------------------------------------
// Setzt den Bereich der Alarmdefinitionen im SRAM auf 0
void DCF_delete_alarm(void)
{
uint8_t i, j;

// die Schaltdefinitionen liegen hinter den Zeit-Arrays !
for (i = CLK_ARRAYS; i < (ALARMDEFS + CLK_ARRAYS); i++)
	{

	for (j = 0; j < TIMEBYTES; j++)
		{
		time[i][j] = 0;
		}
	}
}

// ----------------------------------------------------------------------------
// Lscht den Bereich des Eeproms, in dem die Alarmdefinitionen abgelegt sind
void DCF_delete_eeprom(void)
{
uint16_t eeprom_adr = EEPROM_START;			// Ab Adresse 8 im Eeprom einlesen

while (eeprom_adr < 512)
	{
	EEARH = (uint8_t) (eeprom_adr / 256);	// Adresse schreiben
	EEARL = (uint8_t) eeprom_adr++;

	EEDR  = 0xFF;									// 0xFF  schreiben

	cli();
	EECR |= (1<<EEMWE);
	EECR |= (1<<EEWE);
	sei();
	}
}

#ifdef LOG_RS232
// ----------------------------------------------------------------------------
// gibt die Zeit der aktuellen Minute aus
void print_datum(void)
{

/*
static uint8_t counter = 0;

put_c(13);

if (counter < 100) put_c('0');
if (counter <  10) put_c('0');
put_s(utoa(counter, msg,10));
put_c(' ');
counter++;
if (counter > 239) counter = 0;				// laufende Zeilennummer ausgeben
*/

put_c(' ');

if (time[CLK][STUNDE] < 10) put_c('0');	// Stunde und  Minuten
put_s(utoa(time[CLK][STUNDE], msg , 10));
put_c(':');
if (time[CLK][MINUTE] < 10) put_c('0');
put_s(utoa(time[CLK][MINUTE], msg, 10));

put_c('-');											// Sommer-/Winterzeit
if (is_summertime(&time[CLK][0])) put_c('s'); else put_c('w');
put_c('-');

if (time[CLK][TAG] < 10) put_c('0');		// Tag, Monat, Jahr
put_s(utoa(time[CLK][TAG],msg, 10 ));
put_c('.');
if (time[CLK][MONAT] < 10) put_c('0');
put_s(utoa(time[CLK][MONAT], msg, 10));
put_c('.');
if (time[CLK][JAHR] < 10) put_c('0');
put_s(utoa(time[CLK][JAHR], msg, 10));

put_c(' ');
}

// ----------------------------------------------------------------------------
// gibt die DCF-Bits ab Bit.15 der nchsten Minute aus
// als 0/1, die logischen Gruppen durch einen '.' getrennt
// positive Parittsprfung wird mit '!' dargestellt,
// Fehler in der Parittsprfung werden mit '?' markiert
void print_dcfbits(void)
{
uint8_t i, z;

// Ankndigungsbits -----------------------------------------------------------
for (i = 15; i < 21; i++)
{
	if (DCF_get_bit(i))
	{
		put_c('1');
		z++;
	}
	else put_c('0');
}
// Paritt

// Trenner
put_c('.');

// Minuten --------------------------------------------------------------------
z = 0;
for (i = 21; i < 28; i++)
{
	if (DCF_get_bit(i))
	{
		put_c('1');
		z++;
	}
	else put_c('0');
}
// Paritt
z += DCF_get_bit(28);
if (z & 1) put_c('?'); else put_c('!');	// Paritt muss immer gerade Zahl sein
// Trenner
put_c('.');

// Stunden --------------------------------------------------------------------
z = 0;
for (i = 29; i < 35; i++)
{
	if (DCF_get_bit(i))
	{
		put_c('1');
		z++;
	}
	else put_c('0');
}
// Paritt
z += DCF_get_bit(35);
if (z & 1) put_c('?'); else put_c('!');
// Trenner
put_c('.');

// der Rest --------------------------------------------------------------------
z = 0;
for (i = 36; i < 58; i++)
{
	if (i == 42) put_c('.');
	if (i == 45) put_c('.');
	if (i == 50) put_c('.');
	if (DCF_get_bit(i))
	{
		put_c('1');
		z++;
	}
	else put_c('0');
}
// Paritt
z += DCF_get_bit(58);
if (z & 1) put_c('?'); else put_c('!');

// Ankndigungsbits im Klartext -----------------------------------------------

if (time[DCF][ZBITS] & ~(0x03))				// wenn Ankndigungsbits im DCF-Signal gesendet wurden
														// die Bits fr Sommer-/Winterzeit ignorieren
	{
	if (!(error))									// nur Anzeigen, wenn das Datum fehlerfrei
		{
		if (time[DCF][ZBITS] & 0x10)	puts_p(PSTR(", A1"));
		if (time[DCF][ZBITS] & 0x08)	puts_p(PSTR(", A2"));
		if (time[DCF][ZBITS] & 0x04)	puts_p(PSTR(", R"));
		}
	}
}


// ----------------------------------------------------------------------------
// gibt die Daten des Alarms aus
// zunchst nur die Bytes (wenn Bit 7 gesetzt ist, dann nur "--"),
// ausser bei time[][MODUS], hier bedeutet BIt.7 eine einmalige Aktion
// danach wird die Aktion  noch eimal in Klartext erlutert.
void print_alarm(uint8_t idx)
{
uint8_t i;

puts_p(PSTR("slot: "));						// lfd Nr. der Alarmdefinition
if ((idx - CLK_ARRAYS) < 10) put_c('0');
put_s(itoa((idx - CLK_ARRAYS), msg,10));  // der Index startet hinter den Zeitdaten

put_s(", rule: ");

for (i = 0; i < TIMEBYTES; i++)				// alle Bytes ausgeben
{
if ((time[idx][i] & 128) && (i < TIMEBYTES - 2)) put_s("---"); // leere Definitionen
else
	{
	if (time[idx][i] < 100) put_c('0');		// dezimale Ausgabe mit fhrenden Leerstellen
	if (time[idx][i] <  10) put_c('0');		// ist besser lesbar
	put_s(itoa(time[idx][i], msg, 10));
	}
put_c('.');											// Trennzeichen
}
														// Erluterungen in Klartext

if ((time[idx][MODUS] & 127) < 4)				// eine Schaltaktion, wenn Modus < 4
	{													// Bit.7 ausblenden, da es die Regel nach Ausfhrung lscht
	puts_p(PSTR(" -> switch: "));
	if (time[idx][SWITCH] <  10) put_c('0');
	put_s(itoa((time[idx][SWITCH] & 0xFE), msg, 10));
	puts_p(PSTR(", action: "));
	if      ((time[idx][MODUS] & 0x03) == 1) puts_p(PSTR("on"));
	else if ((time[idx][MODUS] & 0x03) == 2) puts_p(PSTR("off"));
	else if ((time[idx][MODUS] & 0x03) == 3) puts_p(PSTR("toggle"));
	}

else													// an TWI-Slave wird eine Meldung verschickt
	{
	puts_p(PSTR(" -> TWIslave: "));
	put_s(itoa(time[idx][SWITCH], msg, 10));
	puts_p(PSTR(", MSG: "));
	put_s(itoa((time[idx][MODUS] & 127), msg, 10));	// Bit.7 ist nicht Teil der Message
	}
if (time[idx][MODUS] & 0x80) puts_p(PSTR(", single action"));
}

#endif

// --- [ eof ] ----------------------------------------------------------------